(*
  JM_SAP.PAS


  Release Notes:

  1.00: 01-NOV-1996 B.McMullin

     Public release.

*)



PROGRAM autop;
(* program to run the basic autopoiesis algorithm *)
(* Based on an original version for prime 550: j mingers 23/9/82 *)

USES
  Graph, crt;

CONST
  rownum            = 20;
  colnum            = 30;

  (* Constants associated with Graphical User Interface. *)
  LogWindowX        = 50;
  LogWindowY        = 300;
  LogWindowWidth    = 500;
  LogWindowHeight   = 150;
  StateWindowX      = 150;
  StateWindowY      = 10;
  StateWindowWidth  = 360;
  StateWindowHeight = 240;
  WindowBoundary    = 2;
  TextHeight        = 10;
  MinLogX           = 2;
  MinLogY           = 2;

  (* Constants associated with msr88. *)
  msr88_range = 2147483646; (* 2^31 - 2 *)
  msr88_A = 16807;
  msr88_M = MSR88_RANGE + 1;
  msr88_Q = 127773;
  msr88_R = 2836;

TYPE
  array6 = ARRAY[1..6] OF integer;
  array8 = ARRAY[1..8] OF integer;
  array40 = ARRAY[1..40] OF integer;
  array1200 = ARRAY[1..1200] OF integer;

  (* Type associated with Graphical User Interface. *)
  Box =
    RECORD
      x1: Integer;
      y1: Integer;
      x2: Integer;
      y2: Integer;
    END;

VAR
  disintl              : Real;
  noofkat              : Integer;

  grid                 : ARRAY[1..rownum,1..colnum] OF integer;
  oldgrid              : ARRAY[1..rownum,1..colnum] OF integer;
  rel                  : ARRAY[-1..1,-1..1] OF integer;
  adjacent             : ARRAY[1..2] OF array8;
  printchr             : packed ARRAY[0..7] OF char;
  answer               : STRING[8];
  bondlist,linklist    : array8;
  opposite             : array8;
  rowno,colno          : array8;
  katpos               : array40;
  holepos,linkpos      : array1200;
  holecnt,linkcnt      : integer;
  i,iterate            : integer;
  randnums             : integer;
  startprint           : integer;
  fullpr               : boolean;
  paniccnt             : integer;

  (* Variables associated with Graphical User Interface. *)
  Quit                 : Boolean;
  Stepping             : Boolean;
  OddIteration         : Boolean;
  GraphOutput          : Boolean;
  LogX, LogY: Integer;
  CellCenterX, CellCenterY,
  CellForegroundX, CellBackgroundX,
  CellForegroundY, CellBackgroundY: integer;

  (* Variable associated with msr88. *)
  msr88_state : LongInt;


PROCEDURE CursorHome;

BEGIN
  GotoXY(1,1);
END;


PROCEDURE CursorDown;

BEGIN
  GotoXY(WhereX, WhereY+1);
END;


PROCEDURE panic;
(*  Normally just a stub ... but stick a breakpoint here if in doubt. *)

BEGIN
  paniccnt := paniccnt + 1;
  halt;
END;

PROCEDURE prntgrid;
VAR
  prntline: packed ARRAY[1..colnum] OF char;
  i,j,k,l: integer;
  (* NB: holecnt and linkcnt are global. *)

BEGIN
  holecnt := 0;
  linkcnt := 0;
  FOR i:=1 TO rownum DO
    BEGIN
      FOR j:=1 TO colnum DO
        BEGIN
          prntline[j] := '!';
          k := grid[i,j];
          IF k>3
            THEN
              BEGIN
                IF k>=13
                  THEN k := 4
                  ELSE k := 5;
              END;
          prntline[j] := printchr[k];
          IF (k = 2) OR (k > 3)
            THEN linkcnt := linkcnt + 1
          ELSE
            IF k = 0
              THEN holecnt := holecnt + 1;
        END;
      writeln(prntline);
    END;

  if holecnt <> linkcnt
    (* OOOPS.... *)
    THEN panic;

  CursorHome;
  CursorDown;
END;


PROCEDURE listholes;
(* Alternate scanning order, to offset any bias. *)

VAR
  i,j,k: integer;

BEGIN

  FOR i:= 1 TO 1200 DO
    holepos[i] := 0;

  k := 1;

  IF OddIteration
    THEN
      BEGIN
        FOR i:=1 TO rownum DO
          FOR j:=1 TO colnum DO
            IF grid[i,j]=0
              THEN
                BEGIN
                  holepos[k] := i;
                  holepos[k+1] := j;
                  k := k+2;
                END;
      END
    ELSE
      BEGIN
        FOR i:=rownum DOWNTO 1 DO
          FOR j:=colnum DOWNTO 1 DO
            IF grid[i,j]=0
              THEN
                BEGIN
                  holepos[k] := i;
                  holepos[k+1] := j;
                  k := k+2;
                END;
      END;

  holecnt := k div 2;

END;


PROCEDURE listlink;
(* Alternate scanning order, to offset any bias. *)

VAR
  i,j,k: integer;

BEGIN

  FOR i:=1 TO 1200 DO
    linkpos[i] := 0;

  k := 1;

  IF OddIteration
    THEN
      BEGIN
        FOR i:=1 TO rownum DO
          FOR j:=1 TO colnum DO
            IF grid[i,j]=2
              THEN
                BEGIN
                  linkpos[k] := i;
                  linkpos[k+1] := j;
                  k := k+2;
                END;
      END
    ELSE
      BEGIN
        FOR i:=rownum DOWNTO 1 DO
          FOR j:=colnum DOWNTO 1 DO
            IF grid[i,j]=2
              THEN
                BEGIN
                  linkpos[k] := i;
                  linkpos[k+1] := j;
                  k := k+2;
                END;
      END;
END;


PROCEDURE listkats;
(* Alternate scanning order, to offset any bias. *)

VAR
  i,j,k: integer;

BEGIN

  FOR i:=1 TO 40 DO
    katpos[i] := 0;

  k := 1;

  IF OddIteration
    THEN
      BEGIN
        FOR i:=1 TO rownum DO
          FOR j:=1 TO colnum DO
            IF grid[i,j]=3
              THEN
                BEGIN
                  katpos[k] := i;
                  katpos[k+1] := j;
                  k := k+2;
                END;
      END
    ELSE
      BEGIN
        FOR i:=rownum DOWNTO 1 DO
          FOR j:=colnum DOWNTO 1 DO
            IF grid[i,j]=3
              THEN
                BEGIN
                  katpos[k] := i;
                  katpos[k+1] := j;
                  k := k+2;
                END;
      END;

END;

(*
  The following section defines the msr88 random number generator and
  utilities based on it.
*)

FUNCTION msr88 : LongInt;
(*
     This is an implementation of the "minimal standard"
     pseudo random number generator defined in:

        Park, S.K
        Miller, K.W.
        "Random Number Generators: Good ones are hard to find."
        CACM, Oct 88, Vol. 31, No. 10, pp. 1192-1201.
*)
VAR
  test, quotient, remainder : LongInt;

BEGIN
  quotient := msr88_state DIV msr88_Q;
  remainder := msr88_state MOD msr88_Q;
  test := (msr88_A * remainder) - (msr88_R * quotient);
  IF (test > 0)
    THEN msr88_state := test
    ELSE msr88_state := (test + msr88_M);
  msr88 := (msr88_state - 1);
END;

FUNCTION pick(i:integer):integer;
(* Obtain random integer in range 1 to i inclusive. *)

VAR
  cutoff,v : LongInt;

BEGIN
  cutoff := (msr88_range DIV i) * i;
  REPEAT
    v := msr88;
  UNTIL (v < cutoff);
(*
  We discard values of v >= cutoff because, in general, the range of
  msr88 will not be an *integral* multiple of the range required; in
  that case, pick() can only properly deliver a uniform (pseudo)
  probability function by discarding the "extra" values covered by
  msr88. The number of such "extra" values will always be less than i.
  So, since i (being Integer) is necessarily much less than msr88_range,
  the probability of having to discard will be small and we will "almost
  always" exit this loop immediately (i.e. after one iteration).
*)
  pick := (v MOD i) + 1;
END;

FUNCTION bernoulli(p : Real) : Boolean;
(*
  Return a sample of a (pseudo) Bernoulli random variable,
  with parameter p.
*)
VAR
  cut : LongInt;

BEGIN
  cut := Round(p * msr88_range);
  bernoulli := (msr88 < cut);
END;

FUNCTION validsq(i,j:integer): boolean;
(*check whether square is on grid*)
BEGIN
  validsq := (i > 0) AND (i <= rownum) AND (j > 0) AND (j <= colnum);
END;

FUNCTION linkok(i,j,p:integer): boolean;
(* Check if link on i,j can be bonded to single bonded link
   at relative position p *)

VAR
  temp, opp: integer;

BEGIN
  temp := grid[i + rowno[p], j + colno[p] ] - 4;
  opp := opposite[p];
  linkok := (temp <> adjacent[1,opp]) AND
            (temp <> adjacent[2,opp]) AND
            (temp <> opp);
END;

PROCEDURE chekangl(VAR drops:integer;
                   i,j:integer);
(* Checks to see if any neighbouring bonds
   can be linked to a free link. *)

VAR
  l,m,n,opp,temp: integer;

BEGIN
  drops := 0;
  FOR n:=1 TO 8 DO
    BEGIN
      IF bondlist[n]=1
        THEN
          BEGIN
            l := i + rowno[n];
            m := j + colno[n];
            temp := grid[l,m] - 4;
            opp := opposite[n];
            IF (temp = adjacent[1,opp]) OR (temp = adjacent[2,opp])
              THEN
                BEGIN
                  bondlist[n] := 0;
                  drops := drops+1;
                END;
          END;
    END;
END;

PROCEDURE cheklang(VAR drops:integer;
                   i,j:integer);
(* Checks to see if any free links can be bonded
   to the newly formed link on i,j  *)

VAR
  n,temp: integer;

BEGIN
  drops := 0;
  FOR n:=1 TO 8 DO
    BEGIN
      IF linklist[n] = 1
        THEN
          BEGIN
            temp := grid[i,j] - 4;
            IF (n=adjacent[1,temp]) OR (n=adjacent[2,temp])
              THEN
                BEGIN
                  drops := drops + 1;
                  linklist[n] := 0;
                END;
          END;
    END;
END;

PROCEDURE formbond(i,j,p:integer);
(* Bond the link on i,j to the link at rel pos p *)

VAR
  k,l,m,temp: integer;

BEGIN
  l := i + rowno[p];
  m := j + colno[p];
  IF (grid[i,j] < 13) AND (grid[l,m] < 13)
    THEN
      BEGIN
        temp := grid[i,j];
        IF temp = 2
          THEN grid[i,j] := 4 + p
          ELSE grid[i,j] := temp + (8 * p);
        temp := grid[l,m];
        IF temp = 2
          THEN
            BEGIN
              grid[l,m] := 4 + opposite[p];
              (* Amend linkpos for the newly bonded link. *)
              k := l;
              WHILE ((linkpos[k] <> l) OR (linkpos[k+1] <> m))
                    AND (k < 199) DO
                k := k + 2;
              IF k < 199
                THEN
                  IF linkpos[k+1] = m
                    THEN linkpos[k] := 9999;
            END
          ELSE grid[l,m] := temp + (8 * opposite[p]);
      END;
END;

PROCEDURE bond(i,j:integer);
(* Takes a link on i,j and tries to bond it
   firstly to a single bonded link and
   then to a free link. *)

VAR
  l,m,n,bondtot,linktot: integer;
  drops,p: integer;

BEGIN
(* Locate neighbouring links[free] and bonds. *)
  FOR n:=1 TO 8 DO
    BEGIN
      bondlist[n] := 0;
      linklist[n] := 0;
    END;
  bondtot := 0;
  linktot := 0;
  FOR n:=1 TO 8 DO
    BEGIN
      l := i + rowno[n];
      m := j + colno[n];
      IF validsq(l,m)
        THEN
          IF grid[l,m] = 2
            THEN
              BEGIN
              (* free link *)
                linktot := linktot + 1;
                linklist[n] := 1;
              END
          ELSE
            IF (grid[l,m] > 4) AND (grid[l,m] < 13)
              THEN
                BEGIN
                  (* Singly bonded link. *)
                  bondtot := bondtot + 1;
                  bondlist[n] := 1;
                END;
    END;

  (* Deal first with singly bonded links,
     then with free links - if any. *)
  IF bondtot > 0
    THEN
      BEGIN
        chekangl(drops, i, j);
        bondtot := bondtot - drops;
      END;
  IF bondtot >= 2
    THEN
      BEGIN
      (* Select 2 of the bonds - this should be random! *)
        p := 1;
        WHILE bondlist[p]=0 DO
          p := p+1;
        formbond(i, j, p);
        p := p + 1;
        WHILE bondlist[p] = 0 DO
          p := p + 1;
        IF linkok(i, j, p)
          THEN
            IF linkok(i + rowno[p], j + colno[p], opposite[p])
              THEN
                formbond(i, j, p);
      END
    ELSE IF bondtot = 1
           THEN
             BEGIN
               (* Form the 1 bond and check the free links. *)
               p := 1;
               WHILE bondlist[p] = 0 DO
                 p := p + 1;
               formbond(i, j, p);
               IF linktot > 0
                 THEN
                   BEGIN
                   (* Select 1 of the free links - should be random! *)
                     cheklang(drops, i, j);
                     linktot := linktot - drops;
                     IF linktot > 0
                       THEN
                         BEGIN
                           p := 1;
                           WHILE linklist[p] = 0 DO
                             p := p + 1;
                           formbond(i, j, p);
                         END;
                   END;
             END
           ELSE IF linktot > 0
                  THEN
                    BEGIN
                    (* No single bonds so bond 1 or 2 of free links. *)
                      p := 1;
                      WHILE linklist[p] = 0 DO
                        p := p + 1;
                      formbond(i, j, p);
                      linklist[p] := 0;
                      linktot := linktot - 1;
                      cheklang (drops, i, j);
                      linktot := linktot - drops;
                      IF linktot > 0
                        THEN
                          BEGIN
                            p := 1;
                            WHILE linklist[p] = 0 DO
                              p := p + 1;
                            formbond(i,j,p);
                          END;
                    END;
END;


PROCEDURE rebond(i, j:integer);
(* Try to rebond links after one has disintegrated.
   Look first at singly bonded, then free links. *)

VAR
  bondlist: array8;
  links: ARRAY[1..3, 1..12] OF integer;
  l, m, n, sub: integer;


PROCEDURE checkbond(n1,n2:integer);
(* Checks if singly bonded link at rel pos n1 can be
   bonded to rel pos n2 *)

VAR
  l, m, n, o, p: integer;

BEGIN
  IF bondlist[n2] = 1
    THEN
      BEGIN
        l := i + rowno[n1];
        m := j + colno[n1];
        n := i + rowno[n2];
        o := j + colno[n2];
        p := rel[n-l,o-m];
        IF linkok(l, m, p)
          THEN
            BEGIN
              p := rel[l-n, m-o];
              IF linkok(n, o, p)
                THEN
                  BEGIN
                    sub := sub + 1;
                    links[1,sub] := n;
                    links[2,sub] := o;
                    links[3,sub] := p;
                  END;
            END;
      END;
END;


PROCEDURE linkbonds;
(* List surrounding singly bonded links and pair as many as pos. *)

VAR
  n, l: integer;

BEGIN
  FOR n:=1 TO 8 DO
    bondlist[n] := 0;
  FOR n:=1 TO 12 DO
    FOR l:=1 TO 3 DO
      links[l,n] := 0;
  FOR n:=1 TO 8 DO
    BEGIN
      l := i + rowno[n];
      m := j + colno[n];
      IF validsq(l,m)
        THEN
          IF (grid[l,m] > 4) AND (grid[l,m] < 13)
            THEN bondlist[n] := 1;
    END;
  sub := 0; (* NB: sub updated inside checkbond... *)
  FOR n:=1 TO 4 DO
    IF bondlist[n]=1
      THEN
        BEGIN
          checkbond(n, adjacent[1,n]);
          checkbond(n, adjacent[2,n]);
          IF n < 4
            THEN checkbond(n, n+1)
            ELSE checkbond(4, 1);
        END;

  (* Now try to form the bonds for each possible pair.
     formbond checks that no links are doubly bonded. *)
  FOR n:=1 TO sub DO
    formbond(links[1,n], links[2,n], links[3,n]);
END;

BEGIN
  (* First bond singly bonded links. *)
  linkbonds;

  (*now deal with any free links*)
  FOR n:=1 TO 8 DO
    BEGIN
      l := i + rowno[n];
      m := j + colno[n];
      IF validsq(l,m)
        THEN
          IF grid[l,m] = 2
            THEN bond(l, m);
    END;

  (* Now deal with free links that became singly bonded. *)
  linkbonds;
END;


PROCEDURE dsplsub(i,j,l,m:integer);
(* Something on i,j displaces substrate on l,m *)

VAR
  holeno, bondno : array6;
  p, q, n, bondcnt, holecnt : integer;

BEGIN
  FOR n:=1 TO 6 DO
    BEGIN
      holeno[n] := 0;
      bondno[n] := 0;
    END;
  bondcnt := 0;
  holecnt := 0;
  FOR n := 1 TO 4 DO
    BEGIN
      p := l + rowno[n];
      q := m + colno[n];
      IF validsq(p,q)
        THEN
          IF (p <> i) OR (q <> j)
            THEN
            (* Check on adjacent square type. *)
              BEGIN
                IF grid[p,q] = 0
                  THEN
                    BEGIN
                      holecnt := holecnt+1;
                      holeno[(2 * holecnt) - 1] := p;
                      holeno[2 * holecnt] := q;
                    END
                  ELSE IF (grid[p,q] <> 1) AND
                          (grid[p,q] <> 2) AND
                          (grid[p,q] <> 3)
                         THEN
                           BEGIN
                             bondcnt := bondcnt + 1;
                             bondno[(2 * bondcnt) - 1] := p;
                             bondno[2 * bondcnt] := q;
                           END;
              END;
    END;
  IF holecnt <> 0
    THEN
    (* Select 1 of the holes and move. *)
      BEGIN
        n := pick(holecnt);
        p := holeno[(2 * n) - 1];
        q := holeno[2 * n];
        grid[p,q] := 1;
        grid[l,m] := 0;
      END
    ELSE IF bondcnt <> 0
           THEN
           (* Select bonded link and try to move thru. *)
             BEGIN
               n := pick(bondcnt);
               p := bondno[(2 * n) - 1];
               q := bondno[2 * n];

               (* Check the square beyond. *)
               IF (p - l) < 0
                 THEN p := p + 1
                 ELSE IF (p - l) > 0
                        THEN p := p + 1;
               IF (q - m) < 0
                 THEN q := q - 1
                 ELSE IF (q - m) > 0
                        THEN q := q + 1;
               IF validsq(p,q)
                 THEN
                   IF grid[p,q] = 0
                     THEN
                       BEGIN
                         grid[p,q] := 1;
                         grid[l,m] := 0;
                       END;
             END;
END;


PROCEDURE dsplink(i, j, l, m : integer);
(* Something on i,j displaces link on l,m *)

VAR
  n, p, q : integer;
  holeno, subno: array6;
  holecnt, subcnt: integer;

BEGIN
  FOR n:=1 TO 6 DO
    BEGIN
      holeno[n] := 0;
      subno[n] := 0;
    END;
  holecnt := 0;
  subcnt := 0;
  FOR n := 1 TO 4 DO
    BEGIN
      p := l + rowno[n];
      q := m + colno[n];
      IF validsq(p,q)
        THEN
          IF (p <> i) OR (q <> j)
            THEN
              IF grid[p,q] = 0
                THEN
                  BEGIN
                    holecnt := holecnt + 1;
                    holeno[(2 * holecnt) - 1] := p;
                    holeno[2 * holecnt] := q;
                  END
                ELSE IF grid[p,q] = 1
                       THEN
                         BEGIN
                           subcnt := subcnt + 1;
                           subno[(2 * subcnt) - 1] := p;
                           subno[(2 * subcnt)] := q;
                         END;
    END;
  IF holecnt > 0
    THEN
      BEGIN
        n := pick(holecnt);
        p := holeno[(2 * n) - 1];
        q := holeno[2 * n];
        grid[p,q] := 2;
        grid[l,m] := 0;
        bond(p,q);
      END
    ELSE IF subcnt > 0
           THEN
             BEGIN
               n := pick(subcnt);
               p := subno[(2 * n) -1];
               q := subno[2 * n];
               dsplsub(l, m, p, q);
               grid[l,m] := grid[p,q];
               grid[p,q] := 2;
               bond(p,q);
             END;
END;

PROCEDURE arrayfil(VAR ina:array8;
                   a, b, c, d, e, f, g, h:integer);
BEGIN
  ina[1] := a;
  ina[2] := b;
  ina[3] := c;
  ina[4] := d;
  ina[5] := e;
  ina[6] := f;
  ina[7] := g;
  ina[8] := h;
END;

FUNCTION GetChar : Char;

VAR
  Ch: Char;

BEGIN
  Ch := ReadKey;
  IF (Ch = #0)
  THEN
    Ch := ReadKey;
  GetChar := Ch;
END;

PROCEDURE DiscardChar;

VAR
  Ch: Char;

BEGIN
  Ch := GetChar;
END;

PROCEDURE tada;
BEGIN

  ClrScr;

  writeln('            *** Simple Pascal Autopoiesis Simulator ***');
  writeln;
  writeln('                  [Release 1.00 01-NOV-1996]');
  writeln;
  writeln('    Based on:');
  writeln('      "Autopoiesis:');
  writeln('       The Organization of Living Systems,');
  writeln('       its Characterization and a Model",');
  write  ('      Francisco J. Varela, Humberto R. Maturana');
  writeln(' and R. Uribe');
  writeln('      BioSystems (5) pages 187-196, 1974');
  writeln;

  writeln('    Original development by:');
  writeln('     John Mingers,');
  writeln('      University of Warwick, Coventry CV4 7AL, UK.');
  writeln('       Internet: orsjm@wbs.warwick.ac.uk');
  writeln;

  writeln('    Turbo Pascal V5.5 Port by:');
  writeln('     Barry McMullin,');
  writeln('      School of Electronic Engineering,');
  writeln('      Dublin City University, Dublin 9, Ireland.');
  writeln('       Internet: McMullin@EEng.DCU.IE');

  writeln;
  writeln;
  writeln('            (Press any key to continue...)');

  DiscardChar;

END;

PROCEDURE ForceUnit;

VAR i, j : integer;

PROCEDURE ForceLink (i, j : integer);

BEGIN
  grid[i,j] := 2;
  bond(i,j);
END;

PROCEDURE ForceHoles (count : integer);

VAR i, j : integer;

BEGIN
  WHILE count > 0
  DO
    BEGIN
      i := pick(rownum);
      j := pick(colnum);
      IF grid[i,j] = 1 THEN
        BEGIN
        grid[i,j] := 0;
        count := count - 1;
        END;
    END;
END;

BEGIN
  i := rownum div 2;
  j := colnum div 2;
  grid[i,j] := 3;
  noofkat := 1;
  ForceLink(i-2, j-1);
  ForceLink(i-2, j  );
  ForceLink(i-2, j+1);
  ForceLink(i+2, j-1);
  ForceLink(i+2, j  );
  ForceLink(i+2, j+1);
  ForceLink(i-1, j-2);
  ForceLink(i  , j-2);
  ForceLink(i+1, j-2);
  ForceLink(i-1, j+2);
  ForceLink(i  , j+2);
  ForceLink(i+1, j+2);
  ForceHoles(12);
END;


PROCEDURE initial;

VAR i,j,k: integer;
  rn: real;
BEGIN
  arrayfil(rowno, 0,-1, 0,+1,+1,-1,-1,+1);
  arrayfil(colno,-1, 0,+1, 0,-1,-1,+1,+1);
  arrayfil(opposite,3,4,1,2,7,8,5,6);
  arrayfil(adjacent[1],5,6,7,8,4,1,2,3); (* Anti-Clockwise *)
  arrayfil(adjacent[2],6,7,8,5,1,2,3,4); (* Clockwise      *)
  printchr[0] := ' ';   (* space *)
  printchr[1] := '.' ;
  printchr[2] := 'o' ;
  printchr[3] := '*';
  printchr[4] := '@' ;
  printchr[5] := '+' ;
  printchr[6] := '/' ;
  printchr[7] := '\' ;
  FOR i:=1 TO rownum DO
    FOR j:=1 TO colnum DO
      grid[i,j] := 1;
  paniccnt := 0;

  (* I presume this is based on Fig 3.,
     p. 194, of the original BioSystems paper. *)
  rel[-1,-1] := 6;
  rel[-1, 0] := 2;
  rel[-1,+1] := 7;

  rel[ 0,-1] := 1;
  rel[ 0, 0] := 0;
  rel[ 0,+1] := 3;

  rel[+1,-1] := 5;
  rel[+1, 0] := 4;
  rel[+1,+1] := 8;

  tada;

  ClrScr;
  CursorHome;

  writeln('Please input initial random number generator state:');
  readln(msr88_state);
  IF (msr88_state < 1) OR (msr88_state > msr88_range)
  THEN
    BEGIN
      WriteLn('([Invalid value! Using default initial state of 42]');
      msr88_state := 1;
    END;
  writeln;

  writeln('Please input disintegration probability:');
  readln(disintl);
  IF (disintl > 1.0) OR (disintl < 0.0)
  THEN
    BEGIN
      WriteLn('[Invalid value! Using default probability of 0.01]');
      disintl := 0.01;
    END;
  writeln;

  writeln('Please input total number of time steps:');
  readln(iterate);
  writeln;

  stepping := TRUE;

  startprint := 1;
  fullpr := false;
  GraphOutput := true;

(* This options are mainly useful only for debugging;
   so they are commented out in the release version.

  writeln('Please input time step when output should start:');
  readln(startprint);
  writeln;

  writeln('Do you want output at each phase of the algorithm (y/n)?');
  readln(answer);
  IF (answer = 'y') or (answer = 'Y')
    THEN fullpr := true
    ELSE fullpr := false;

  writeln;
  writeln('Do you want graphical (EGA/VGA, Colour) output (y/n)?');
  readln(answer);
  IF (answer = 'y') or (answer = 'Y')
    THEN GraphOutput := true
    ELSE GraphOutput := false;
*)

  writeln('Do you want a pregenerated unit (y/n)?');
  readln(answer);
  IF (answer = 'y') or (answer = 'Y')
    THEN ForceUnit
    ELSE
      BEGIN
        writeln;
        writeln('Please input number of catalysts (must be <= 20):');
        readln(noofkat);
        IF (noofkat > 20) OR (noofkat < 0)
        THEN
          BEGIN
            WriteLn('[Invalid value! Using default of 1 catalyst.]');
            noofkat := 1;
          END;
        FOR k:=1 TO noofkat DO
          BEGIN
            REPEAT (* Ensure we get a *new* catalyst every time. *)
              i := pick(rownum);
              j := pick(colnum);
            UNTIL grid[i,j] <> 3;
            grid[i,j] := 3;
          END;
      END;
END;

PROCEDURE movehole;
(* Make any movements into holes. *)

VAR
  i,j,k,l,m,rn : integer;

BEGIN
  listholes;
  k := 1;
  WHILE holepos[k] <> 0 DO
    BEGIN
      rn := pick(4);
      i := holepos[k];
      j := holepos[k+1];
      l := i + rowno[rn];
      m := j + colno[rn];
      IF grid[i,j] <> 0 THEN panic;
      IF validsq(l,m)
        THEN
          BEGIN
            IF (grid[l,m] = 1) OR (grid[l,m] = 3)
              THEN
                BEGIN
                  grid[i,j] := grid[l,m];
                  grid[l,m] := 0;
                END
              ELSE IF grid[l,m] = 2
                     THEN
                       BEGIN
                         grid[i,j] := 2;
                         grid[l,m] := 0;
                         bond(i,j);
                       END
                     ELSE IF grid[l,m] <> 0
                            THEN
                              BEGIN
                              (* Move thru bonded links if a substrate. *)
                                CASE rn OF
                                  1: m := m - 1;
                                  2: l := l - 1;
                                  3: m := m + 1;
                                  4: l := l + 1;
                                END;
                                IF validsq(l,m)
                                  THEN
                                    IF grid[l,m] = 1
                                      THEN
                                        BEGIN
                                          grid[i,j] := 1;
                                          grid[l,m] := 0;
                                        END;
                              END;
          END;
      k := k + 2;
    END;
END;

PROCEDURE movelink;

VAR
  i,j,k,l,m,rn  : integer;

BEGIN
  listlink;
  k := 1;
  WHILE linkpos[k] <> 0 DO
    BEGIN
      IF linkpos[k] <> 9999
        THEN
          BEGIN
            rn := pick(4);
            i := linkpos[k];
            j := linkpos[k+1];
            l := i+rowno[rn];
            m := j+colno[rn];
            IF (grid[i,j] = 2) AND validsq(l,m)
              THEN
                IF grid[l,m] = 0
                  THEN
                    BEGIN
                      grid[l,m] := 2;
                      grid[i,j] := 0;
                      bond(l,m);
                    END
                  ELSE IF grid[l,m] = 1
                         THEN
                           BEGIN
                             dsplsub(i,j,l,m);
                             grid[i,j] := grid[l,m];
                             grid[l,m] := 2;
                             bond(l,m);
                           END;
          END;
      k := k + 2;
    END;
END;

PROCEDURE movekats;

VAR
  temp,i,j,k,l,m,rn : integer;

BEGIN
  listkats;
  k := 1;
  WHILE katpos[k] <> 0 DO
    BEGIN
      rn := pick(4);
      i := katpos[k];
      j := katpos[k+1];
      l := i + rowno[rn];
      m := j + colno[rn];
      IF validsq(l,m)
        THEN
          IF grid[l,m] = 0
            THEN
              BEGIN
                grid[l,m] := 3;
                grid[i,j] := 0;
              END
            ELSE IF grid[l,m] = 1
                   THEN
                     BEGIN
                       dsplsub(i,j,l,m);
                       temp := grid[i,j];
                       grid[i,j] := grid[l,m];
                       grid[l,m] := temp;
                     END
                   ELSE IF grid[l,m] = 2
                          THEN
                            BEGIN
                              dsplink(i,j,l,m);
                              temp := grid[i,j];
                              grid[i,j] := grid[l,m];
                              grid[l,m] := temp;
                              IF grid[i,j] = 2
                                THEN bond(i,j);
                            END;
      k := k + 2;
    END;
END;

PROCEDURE produce;

VAR
  i,j,k,l,m,n,p,q,rn: integer;
  subs: array8;

BEGIN
  listkats;
  k := 1;
  WHILE katpos[k] <> 0 DO
    BEGIN
      FOR n:=1 TO 8 DO
        subs[n] := 0;
      i := katpos[k];
      j := katpos[k+1];
      FOR n:=1 TO 8 DO
      (* Locate neighbouring substrates. *)
        BEGIN
          l := i + rowno[n];
          m := j + colno[n];
          IF validsq(l,m)
            THEN
              IF grid[l,m] = 1
                THEN subs[n] := 1;
        END;
      p := 0;
      FOR n:=1 TO 8 DO
      (* Check on substrates' neighbours. *)
        IF subs[n]=1
          THEN
            BEGIN
              p := p + 1;
              l := adjacent[1,n];
              m := adjacent[2,n];
              IF (subs[l] = 0) AND (subs[m] = 0)
                THEN
                (* No neighbouring substrate so delete. *)
                  BEGIN
                    subs[n] := 0;
                    p := p - 1;
                  END;
            END;
      IF p > 0
        THEN
        (* Randomly choose 1 to produce new link. *)
          BEGIN
            rn := pick(p);
            q := 0;
            n := 0;
            REPEAT
              n := n + 1;
              WHILE subs[n]=0 DO
                n := n + 1;
              q := q + 1;
            UNTIL q = rn;
            (* Create link. *)
            l := i + rowno[n];
            m := j + colno[n];
            grid[l,m] := 2;
            bond(l,m);
            (* Create hole. *)
            l := adjacent[1,n];
            m := adjacent[2,n];
            IF (subs[l] = 1) AND (subs[m] = 1)
              THEN
              (* Both neighbours are substrates so choose. *)
                BEGIN
                  rn := pick(2);
                  IF rn = 1
                    THEN n := l
                    ELSE n := m;
                END
              ELSE IF subs[l] = 1
                     THEN n := l
                     ELSE n := m;
            l := i + rowno[n];
            m := j + colno[n];
            grid[l,m] := 0;
          END;
      k := k + 2;
    END;
END;

PROCEDURE disint;

VAR
  i,j,n: integer;
  k,l,temp,temp1: integer;
  spacefound: boolean;

PROCEDURE debond;
BEGIN
  k := i + rowno[temp];
  l := j + colno[temp];
  IF grid[k,l] < 13
    THEN grid[k,l] := 2
    (* Other bond was only single *)
    ELSE
      BEGIN
        temp1 := (grid[k,l] - 5) DIV (8);
        IF temp1 = opposite[temp]
          THEN
            grid[k,l] := grid[k,l] - (8 * temp1)
          ELSE grid[k,l] := 4 + temp1;
      END;
END;

BEGIN
  FOR i:=1 TO rownum DO
    FOR j:=1 TO colnum DO
      BEGIN
        IF (grid[i,j] = 2) OR (grid[i,j] >= 4)
          THEN
            BEGIN
              IF bernoulli(disintl)
                THEN
                  BEGIN
                    IF grid[i,j] = 2
                      THEN grid[i,j] := 1
                      ELSE
                        BEGIN
                          IF grid[i,j] >= 13
                            THEN
                              BEGIN
                                temp := (grid[i,j] - 5) DIV (8);
                                grid[i,j] := grid[i,j] - (8 * temp);
                                debond;
                              END;
                          IF grid[i,j] >= 4
                            THEN
                              BEGIN
                                temp := grid[i,j] - 4;
                                grid[i,j] := 1;
                                debond;
                              END;
                          rebond(i,j);
                        END;
                    n := 0;
                    spacefound := false;

                    REPEAT
                      n := n + 1;
                      k := i + rowno[n];
                      l := j + colno[n];
                      IF validsq(k,l)
                        THEN
                          IF grid[k,l]=0
                            THEN
                              BEGIN
                                grid[k,l] := 1;
                                spacefound := true;
                              END;
                    UNTIL (spacefound) OR (n = 8);
                     (* Check Moore (8-cell) neighbourhood. *)

                    IF NOT spacefound
                      (* No immediate neighbour hole.
                         Just pick *some* hole (at random)...
                        *)
                      THEN
                        BEGIN
                          listholes;
                          IF (holecnt = 0)
                            THEN panic;
                          n := (2 * pick(holecnt)) - 1;
                          k := holepos[n];
                          l := holepos[n+1];
                          IF (grid[k,l] <> 0)
                            THEN panic;
                          grid[k,l] := 1;
                          spacefound := TRUE;
                        END;

                  END;
            END;
      END;
END;

PROCEDURE ConditionalPause;

VAR
  ch: Char;

BEGIN
  IF stepping
  THEN
    BEGIN
      Ch := GetChar;
      CASE Ch OF
        'g' : stepping := FALSE;
      END;
    END
  ELSE
    BEGIN
      IF KeyPressed
      THEN
        BEGIN
        DiscardChar; (* Discard the interrupt char... *)
        Ch := GetChar;
        stepping := (Ch = ' ');
        END;
  END;
END;



PROCEDURE TextFrontEnd;


BEGIN
  ClrScr;
  writeln(' Initial position:');
  prntgrid;

  OddIteration := TRUE;
  FOR i:=1 TO iterate DO
    BEGIN
      ConditionalPause;
      CursorHome;
      IF i >= startprint
        THEN writeln('        Time Step: ',i);
      movehole;
      IF (i >= startprint) AND (fullpr)
        THEN prntgrid;
      movelink;
      IF (i >= startprint) AND (fullpr)
        THEN prntgrid;
      movekats;
      IF (i >= startprint) AND (fullpr)
        THEN prntgrid;
      produce;
      IF (i >= startprint) AND (fullpr)
        THEN prntgrid;
      disint;
      IF i >= startprint
        THEN prntgrid;

      OddIteration := (NOT OddIteration);
    END;

  WriteLn(' Run Ended.');
  WriteLn(' Press any key to Exit.');
  DiscardChar;

END;


(* Graphical Front-end procedures follow. *)

PROCEDURE InitialiseDisplay;

VAR
  GraphDriver, GraphMode, GraphCode: Integer;
BEGIN
  GraphDriver := Detect;
  InitGraph(GraphDriver,GraphMode,'');
  GraphCode := GraphResult;
  IF GraphCode <> grOK
    THEN
      BEGIN
        WriteLn('Graphics error: ',
                GraphErrorMsg(GraphCode));
        Halt(1);
      END;


  CellCenterX     := StateWindowWidth DIV (4 * colnum);
  CellForegroundX := 2 * CellCenterX;
  CellBackgroundX := 2 * CellForegroundX;

  CellCenterY     := StateWindowHeight DIV (4 * rownum);
  CellForegroundY := 2 * CellCenterY;
  CellBackgroundY := CellForegroundY * 2;

  LogX := MinLogX;
  LogY := MinLogY;
  MoveTo(StateWindowX, StateWindowY - TextHeight);
  OutText('State');
  rectangle(
            StateWindowX-WindowBoundary,
            StateWindowY-WindowBoundary,
            StateWindowX+StateWindowWidth+WindowBoundary,
            StateWindowY+StateWindowHeight+WindowBoundary);
  MoveTo(LogWindowX, LogWindowY - TextHeight);
  OutText('Data Logger');
  rectangle(
            LogWindowX-WindowBoundary,
            LogWindowY-WindowBoundary,
            LogWindowX+LogWindowWidth+WindowBoundary,
            LogWindowY+LogWindowHeight+WindowBoundary);
END;

PROCEDURE SetStateWindow;
BEGIN
  SetViewPort(
              StateWindowX, StateWindowY,
              StateWindowX+StateWindowWidth, StateWindowY+StateWindowHeight,
              true);
END;

PROCEDURE SetLogWindow;
BEGIN
  SetViewPort(
              LogWindowX, LogWindowY,
              LogWindowX+LogWindowWidth, LogWindowY+LogWindowHeight,
              true);
  SetColor(White);
END;

PROCEDURE GrWriteStr(s: STRING);
BEGIN
  SetLogWindow;
  MoveTo(LogX, LogY);
  OutText(s);
  LogX := GetX;
  LogY := GetY;
END;

PROCEDURE GrWriteInt(i : integer);

VAR
  s : STRING[10];
BEGIN
  Str(i, s);
  GrWriteStr(s);
END;

PROCEDURE GrWriteLn;
BEGIN
  LogX := MinLogX;
  LogY := LogY + TextHeight;
  IF LogY > (LogWindowHeight - TextHeight - MinLogY)
    THEN
      LogY := MinLogY;
  SetFillStyle(SolidFill,Black);
  bar(LogX, LogY, LogWindowWidth, LogY+TextHeight-1);
END;

PROCEDURE SetBox(VAR r: Box;
                       x1: Integer; y1: Integer;
                       x2: Integer; y2: Integer);
BEGIN
  r.x1 := x1;
  r.y1 := y1;
  r.x2 := x2;
  r.y2 := y2;
END;

PROCEDURE EraseStateWindow;
BEGIN
  SetStateWindow;
  SetFillStyle(EmptyFill,Black);
  bar(1,1,StateWindowWidth,StateWindowHeight);
END;

PROCEDURE EraseBox(r: Box);
BEGIN
  SetFillStyle(EmptyFill,Black);
  bar(r.x1, r.y1, r.x2, r.y2);
END;

PROCEDURE FrameRedBox(r: Box);
BEGIN
  SetColor(LightRed);
  r.x1 := r.x1 + 1;
  r.y1 := r.y1 + 1;
  r.x2 := r.x2 - 1;
  r.y2 := r.y2 - 1;
  rectangle(r.x1, r.y1, r.x2, r.y2);
END;

PROCEDURE FrameBlueBox(r: Box);
BEGIN
  SetColor(LightBlue);
  rectangle(r.x1, r.y1, r.x2, r.y2);
END;

PROCEDURE PaintYellowBox(r: Box);
BEGIN
  SetFillStyle(SolidFill,Yellow);
  bar(r.x1, r.y1, r.x2, r.y2);
END;

PROCEDURE BlueForay(x, y : Integer);
BEGIN
  SetColor(LightBlue);
  LineRel(x, y);
  MoveRel(-x, -y);
END;

PROCEDURE BlackForay(x, y : Integer);
BEGIN
  SetColor(Black);
  LineRel(x, y);
  MoveRel(-x, -y);
END;

PROCEDURE EraseCell(i, j, element : Integer);

VAR
  r: Box;
  xi, yi : integer;
  BondDirection : integer;

BEGIN
  IF (element <> 0) THEN
    BEGIN
      r.x1 := ((j - 1) * CellBackgroundX) + 1;
      r.y1 := ((i - 1) * CellBackgroundY) + 1;
      r.x2 := r.x1 + CellBackgroundX - 3;
      r.y2 := r.y1 + CellBackgroundY - 3;
      IF element > 3 (* Bond(s) present *)
        THEN
          BEGIN
            r.x1 := r.x1 - 1;
            r.y1 := r.y1 - 1;
            r.x2 := r.x2 + 1;
            r.y2 := r.y2 + 1;
          END;
      EraseBox(r);
    END;
END;


PROCEDURE DisplayCell(i, j, element : Integer);

VAR
  r: Box;
  BondDirection : integer;
  xi, yi : integer;

BEGIN

  xi := ((j - 1) * CellBackgroundX) + CellCenterX;
  yi := ((i - 1) * CellBackgroundY) + CellCenterY;

  SetBox(r, xi, yi,
         xi + CellForegroundX - 1,
         yi + CellForegroundY - 1);
  IF element = 1 (* Substrate *)
    THEN FrameRedBox(r)
    ELSE
      IF element = 3 (* Catalyst *)
        THEN PaintYellowBox(r)
        ELSE
          IF element <> 0 (* Must be some kind of Link *)
            THEN
              BEGIN
                FrameBlueBox(r);
                IF element > 3 (* Bond(s) present *)
                  THEN
                    BEGIN
                      MoveTo(xi + CellCenterX, yi + CellCenterY);
                      IF element > 13
                        THEN
                          BEGIN
                            BondDirection := (element - 5) DIV 8;
                            element := element - (BondDirection * 8);
                            BlueForay(
                              CellBackgroundX * colno[BondDirection],
                              CellBackgroundY * rowno[BondDirection]);
                          END;
                      IF element > 3
                        THEN
                          BEGIN
                            BondDirection := element - 4;
                            BlueForay(
                              CellBackgroundX * colno[BondDirection],
                              CellBackgroundY * rowno[BondDirection]);
                          END;
                    END;
              END;
END;



PROCEDURE DisplayGrid;

VAR
  i, j: integer;
  element : integer;

BEGIN
  SetStateWindow;
  holecnt := 0;
  linkcnt := 0;
  FOR i:=1 TO rownum DO
    BEGIN
      FOR j:=1 TO colnum DO
        BEGIN
          element := grid[i,j];
          IF (element = 2) OR (element > 3)
            THEN linkcnt := linkcnt + 1
            ELSE
              IF element = 0
                THEN holecnt := holecnt + 1;
          oldgrid[i,j] := element;
          DisplayCell(i, j, element);
        END;
    END;
  IF holecnt <> linkcnt
    THEN panic;

END;

PROCEDURE FullRefreshGrid;
BEGIN
  EraseStateWindow;
  DisplayGrid;
END;

PROCEDURE IncrementalRefreshGrid;

VAR
  i, j: integer;
  element, oldelement : integer;

BEGIN
  SetStateWindow;
  holecnt := 0;
  linkcnt := 0;
  FOR i:=1 TO rownum DO
    BEGIN
      FOR j:=1 TO colnum DO
        BEGIN
          element := grid[i,j];
          IF (element = 2) OR (element > 3)
            THEN linkcnt := linkcnt + 1
            ELSE
              IF element = 0
                THEN holecnt := holecnt + 1;
          oldelement := oldgrid[i,j];
          IF (element <> oldelement)
            THEN EraseCell(i, j, oldelement);
        END;
    END;
  IF holecnt <> linkcnt
    THEN panic;

  FOR i:=1 TO rownum DO
    BEGIN
      FOR j:=1 TO colnum DO
        BEGIN
          element := grid[i,j];
          oldelement := oldgrid[i,j];
          IF (element <> oldelement)
            THEN
              BEGIN
                oldgrid[i,j] := element;
                DisplayCell(i, j, element);
              END;
        END;
    END;

END;

PROCEDURE GrProcessChar;

VAR
  Ch : Char;

BEGIN
  Ch := GetChar;
  CASE Ch OF
    'r' :
      BEGIN
        FullRefreshGrid;
        GrProcessChar; (* Recursive! *)
      END;
    'q' : Quit := TRUE;
    'g' : stepping := FALSE;
  END;
END;

PROCEDURE GrConditionalPause;

VAR
  ch: Char;

BEGIN
  IF stepping
  THEN
    GrProcessChar
  ELSE
    BEGIN
      IF KeyPressed
      THEN
        BEGIN
          DiscardChar; (* Discard the interrupt char... *)
          GrProcessChar;
        END;
  END;
END;


PROCEDURE GraphicalFrontEnd;

BEGIN
  InitialiseDisplay;
  GrWriteStr(' Initial position.');
  GrWriteLn;
  DisplayGrid;

  OddIteration := TRUE;
  Quit := FALSE;
  i := 1;
  WHILE (i <= iterate) AND (NOT Quit) DO
    BEGIN
      GrConditionalPause;
      IF (NOT Quit) THEN
        BEGIN
          IF i>=startprint
            THEN
             begin
              GrWriteStr(' Time Step: ');
              GrWriteInt(i);
              GrWriteLn;
             end;
          movehole;
          IF (i >= startprint) AND (fullpr)
            THEN IncrementalRefreshGrid;
          movelink;
          IF (i >= startprint) AND (fullpr)
            THEN IncrementalRefreshGrid;
          movekats;
          IF (i >= startprint) AND (fullpr)
            THEN IncrementalRefreshGrid;
          produce;
          IF (i >= startprint) AND (fullpr)
            THEN IncrementalRefreshGrid;
          disint;
          IF i >= startprint
            THEN IncrementalRefreshGrid;

          OddIteration := (NOT OddIteration);
          i := i + 1;
        END;
    END;

  GrWriteStr(' Run Ended.');
  GrWriteLn;
  GrWriteStr(' Press any key to Exit.');
  GrWriteLn;
  DiscardChar;
  CloseGraph;
END;

BEGIN

  initial;

  IF GraphOutput
    THEN GraphicalFrontEnd
    ELSE TextFrontEnd;

END.
